/* xroute.c, Copyright (c) 1993, 1994, by Kartik Subbarao

 Permission is granted, without charge, to use, copy, modify and distribute 
 this software for any purpose, provided that this copyright notice appears 
 prominently in all copies.

 This software is provided AS IS, without warranty of ANY KIND, expressed
 or implied, including but not limited to the warranties of merchantability,
 fitness for a particular purpose and noninfringement. In NO EVENT shall the 
 author be liable for ANY claim in connection with the use of this software.

 */

# include <stdio.h>
# include <stdlib.h>
# include <sys/types.h>
# include <sys/time.h>
# include <netdb.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <sys/socket.h>
# include <string.h>
# include <sys/wait.h>
# include <values.h>
#ifdef sgi
# define _BSD_SIGNALS
#endif
# include <signal.h>
# include <errno.h>

extern char *sys_errlist[]; extern int errno; 

# define XPORT 6000
# define ROUTEPORT 5500

# define POPUPCMD "/usr/bin/X11/xdialog -d %s:%d -t xroute -O Yes -C No -m 'Accept X connection from %s?'"
/* An alternative, if you don't have xdialog */
/* # define POPUPCMD "/usr/bin/X11/xmessage -display %s:%d -buttons Yes:0,No:1 Accept X connection from %s?" */

usage()
{
	fprintf(stderr, "usage: xroute [-i] [-r routehost] [-n numclients] [-d gdisp] [-q] host[:display]\n");
	exit(1);
}

#if defined(hpux) || defined(linux)
waiter() {int st; while (wait3(&st,WNOHANG,NULL) > 0); signal(SIGCHLD,waiter);}
#else
waiter() { union wait st; while (wait3(&st, WNOHANG, NULL) > 0); }
#endif

/* administrative limits */
int stdisp = 0, endisp = MAXINT-XPORT, maxcli = MAXINT; 

main(argc, argv)
int argc; char *argv[];
{
	int i, o; extern char *optarg; extern int optind;
	char *xhost, *p;
	int xdisp, gdisp = -1; /* Display numbers */
	int xs, gs, os; /* X socket, gateway socket, other socket */
	int inetd = 0; /* Run from inetd? */
	int n, numcli = 1; /* Number of clients */
	int query = 0; /* Query on accepting incoming connections? */
	char buf[128], *av[128], t; int ac;

opts:
	optind = 1;
	while ((o = getopt(argc, argv, "in:d:qr:")) != -1)
		switch (o) {
		case 'i': if (inetd == 0) inetd = 1; break;
		case 'n':
			n = atoi(optarg);
			if (inetd == 1) maxcli = n; else numcli = n;
			break;
		case 'd':
			if (inetd == 1) {
				stdisp = atoi(optarg);
				if (p = strchr(optarg, '-')) endisp = atoi(p+1);
			}
			else gdisp = atoi(optarg); 
			break;
		case 'q': query = 1; break;
		case 'r':
			gs = tcpcsocket(optarg, ROUTEPORT);
			/* write remaining command line arguments to remote host */
			for (i = optind; i < argc; i++) {
				write(gs, argv[i], strlen(argv[i]));
				if (i != argc-1) write(gs, " ", 1);
			}
			write(gs, "\n", 1);

			/* print any output from the remote host */
			while ((n = read(gs, buf, sizeof buf)) > 0) write(1, buf, n);
			exit(0);
		default: usage();
		}

	if (inetd == 1) { /* Run from inetd on the gateway machine */
		fgets(buf, sizeof buf, stdin);
		for (ac = 1, av[1] = p = buf; *p; p++) {
			t = *p;
			if (t == ' ' || t == '\n') {
				*p = '\0'; 
				av[++ac] = p+1; 
				if (t == '\n') break;
			}
		}
		argv = av; argc = ac;
		inetd = 2; goto opts; /* Parse user's arguments */
	}

	if (numcli > maxcli) {
		printf("%d clients requested, allowing only %d\n", numcli, maxcli);
		numcli = maxcli;
	}

	if (!(xhost = argv[optind]))
		if (!(xhost = getenv("DISPLAY"))) usage();

	xdisp = (p = strchr(xhost, ':')) && p[1] ? atoi(p+1) : 0;
	if (p) *p = '\0';

	gs = xsocket(gdisp);
	signal(SIGCHLD, waiter);
	signal(SIGHUP, SIG_IGN);

	if (fork()) exit(0); /* background ourself */
	if (inetd) close(0), close(1), close(2);

	for (; numcli; numcli--) {
		struct sockaddr_in rad; int radlen = sizeof rad;

		while ((os = accept(gs, &rad, &radlen)) == -1 && errno == EINTR);
		if (os == -1) perror("accept"), exit(1);

		if (query) {
			struct hostent *he;
			char cmd[256], *peer;

			he = gethostbyaddr(&rad.sin_addr, sizeof(rad.sin_addr), AF_INET);
			peer = he ?	he->h_name : inet_ntoa(rad.sin_addr);

			sprintf(cmd, POPUPCMD, xhost, xdisp, peer);
			if (system(cmd)) { close(os); continue; }
		}

		switch (fork()) {
		case -1: exit(1);
		case 0: 
			close(gs); 
			xs = tcpcsocket(xhost, XPORT + xdisp); route(xs, os); 
			break;
		default: close(os);
		}
	}
	exit(0);
}

route(xs, os)
int xs, os;
{
	int maxfds, xn = 0, on = 0, tmp; 
	fd_set rdset, wrset, saveset;
	char xobuf[4096], *xop, oxbuf[4096], *oxp;
	
	maxfds = (xs > os ? xs : os) + 1;
	FD_ZERO(&saveset); FD_SET(xs, &saveset); FD_SET(os, &saveset);
	xop = xobuf; oxp = oxbuf;

	for (;;) {
		if (xn == 0 && on == 0) {
			rdset = saveset;
			if (select(maxfds, &rdset, NULL, NULL, NULL) == -1) exit(1);
			if (FD_ISSET(xs, &rdset))
				if ((xn = read(xs, xobuf, sizeof xobuf)) <= 0) exit(xn);
			if (FD_ISSET(os, &rdset))
				if ((on = read(os, oxbuf, sizeof oxbuf)) <= 0) exit(on);
		}
		else {
			if (xn) FD_SET(xs, &rdset);
			if (on) FD_SET(os, &rdset);
		}
	
		wrset = saveset; 
		if (select(maxfds, NULL, &wrset, NULL, NULL) == -1) exit(1);

		if (FD_ISSET(xs, &rdset) && FD_ISSET(os, &wrset)) {
			if ((tmp = write(os, xop, xn)) <= 0) exit(tmp);
			if (xn -= tmp) xop += tmp; 
			else { FD_CLR(xs, &rdset); xop = xobuf; xn = 0; }
		}

		if (FD_ISSET(os, &rdset) && FD_ISSET(xs, &wrset)) {
			if ((tmp = write(xs, oxp, on)) <= 0) exit(tmp);
			if (on -= tmp) oxp += tmp;
			else { FD_CLR(os, &rdset); oxp = oxbuf; on = 0; }
		}
	}
}

int xsocket(gdisp)
int gdisp;
{
	int i, s; struct sockaddr_in sa;
   	 
    if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		perror("socket"), exit(1);

   	sa.sin_family = AF_INET; sa.sin_addr.s_addr = htonl(INADDR_ANY);

	if (gdisp == -1) { /* If it wasn't specified */
		/* find an acceptable port */
		for (i = XPORT+stdisp; i <= XPORT+endisp; i++) {
			sa.sin_port = htons(i);
			if (bind(s, &sa, sizeof(sa)) == 0) {
				printf("%d\n", i-XPORT); break;
			}
			else printf("bind: %d: %s\n", i, sys_errlist[errno]);
		}

		if (i > XPORT+endisp) {
			printf("All allowed X ports in use\n");
			exit(1);
		}
	}
	else {
		if (stdisp && gdisp < stdisp) {
			printf("display %d below minimum allowed %d\n", gdisp, stdisp);
			exit(1);
		}
		if (endisp != MAXINT-XPORT && gdisp > endisp) {
			printf("display %d above maximum allowed %d\n", gdisp, endisp);
			exit(1);	
		}

		sa.sin_port = htons(XPORT+gdisp); 
		if (bind(s, &sa, sizeof(sa)) == -1) perror("bind"), exit(1);
	}

    if (listen(s, 5) == -1) perror("listen"), exit(1);

	return s;
}

int tcpcsocket(host, port)
char *host; int port;
{
	int s, e;
	struct sockaddr_in sa;
	struct hostent *hent;

	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) exit(1);

	if (! (hent = gethostbyname(host))) {
		e = errno;
		if ((sa.sin_addr.s_addr = inet_addr(host)) == -1)
			fprintf(stderr, "%s: %s\n", host, sys_errlist[e]), exit(1);
	}
   	else bcopy(hent->h_addr, (char *) &sa.sin_addr, hent->h_length);

	sa.sin_family = AF_INET; sa.sin_port = htons(port);

	if (connect(s, &sa, sizeof(sa)) == -1)
		fprintf(stderr, "%s: connect: %s\n", host, sys_errlist[errno]), exit(1);

	return s;
}
